/* Generated by Cython 0.29.14 */

#ifndef __PYX_HAVE__pycalphad__core__composition_set
#define __PYX_HAVE__pycalphad__core__composition_set

#include "Python.h"
struct CompositionSetObject;

/* "pycalphad/core/composition_set.pxd":4
 * from pycalphad.core.phase_rec cimport PhaseRecord
 * 
 * cdef public class CompositionSet(object)[type CompositionSetType, object CompositionSetObject]:             # <<<<<<<<<<<<<<
 *     cdef public PhaseRecord phase_record
 *     cdef readonly double[::1] dof, X
 */
struct CompositionSetObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_9pycalphad_4core_15composition_set_CompositionSet *__pyx_vtab;
  struct PhaseRecordObject *phase_record;
  __Pyx_memviewslice dof;
  __Pyx_memviewslice X;
  __Pyx_memviewslice _X_2d_view;
  double NP;
  int zero_seen;
  double energy;
  __Pyx_memviewslice _energy_2d_view;
  __Pyx_memviewslice grad;
  __Pyx_memviewslice _prev_dof;
  double _prev_energy;
  __Pyx_memviewslice _prev_grad;
  int _first_iteration;
};

#ifndef __PYX_HAVE_API__pycalphad__core__composition_set

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C DL_IMPORT(PyTypeObject) CompositionSetType;

#endif /* !__PYX_HAVE_API__pycalphad__core__composition_set */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcomposition_set(void);
#else
PyMODINIT_FUNC PyInit_composition_set(void);
#endif

#endif /* !__PYX_HAVE__pycalphad__core__composition_set */
